import argparse
import textwrap
from unipoll_api.config import get_settings

settings = get_settings()


# Check if IP address is valid
def check_ip(arg_value):
    address = arg_value.split(".")
    if len(address) != 4:
        raise argparse.ArgumentTypeError("invalid host value")
    for i in address:
        if int(i) > 255 or int(i) < 0:
            raise argparse.ArgumentTypeError("invalid host value")
    return arg_value


# Parse CLI arguments
def parse_args():
    parser = argparse.ArgumentParser(description="University Polling API")
    subparser = parser.add_subparsers(title='Available commands', dest='command', required=True)

    run_parser = subparser.add_parser('run',
                                      help="Run the API server",
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      description=textwrap.dedent('''\
        Run University Polling API
        --------------------------------
        Examples:

            python main.py --reload --host=127.0.0.1 --port=8000
            python main.py --reload
        '''))

    run_parser.add_argument("--reload", action="store_true", help="Enable auto-reload", default=False)
    run_parser.add_argument("--host", type=check_ip, default=settings.host, help="Host IP address")
    run_parser.add_argument("--port", type=int, default=settings.port, help="Host port number")

    subparser.add_parser('setup', help="Setup the API server")
    subparser.add_parser('get-openapi', help="Get the OpenAPI schema")

    return parser.parse_args()



from unipoll_api.documents import ResourceID, Account, Group
from unipoll_api.exceptions import resource


# Exception for when a Group with the same name already exists
class NonUniqueName(resource.NonUniqueName):
    def __init__(self, group: Group):
        super().__init__("Group", resource_name=group.name)


# Exception for when an error occurs during Group creation
class ErrorWhileCreating(resource.ErrorWhileCreating):
    def __init__(self, group: Group):
        super().__init__("Group", resource_name=group.name)


# Exception for when a Group is not found
class GroupNotFound(resource.ResourceNotFound):
    def __init__(self, group_id: ResourceID):
        super().__init__("Group", resource_id=group_id)


# Not authorized
class UserNotAuthorized(resource.UserNotAuthorized):
    def __init__(self, account: Account, group: Group, action: str):
        super().__init__(account, f'group {group.name}', action)


# Exception for when a Group was not deleted successfully
class ErrorWhileDeleting(resource.ErrorWhileDeleting):
    def __init__(self, group_id: ResourceID):
        super().__init__("Workspace", resource_id=group_id)


# Exception for trying to add a member that already exists
class AddingExistingMember(resource.AddingExistingMember):
    def __init__(self, group: Group, user: Account):
        super().__init__(group, user)


# Action not found
class ActionNotFound(resource.ActionNotFound):
    def __init__(self, action: str):
        super().__init__('group', action)


# User is not a member of the group
class UserNotMember(resource.UserNotMember):
    def __init__(self, group: Group, user: Account):
        super().__init__(group, user)


# Error while removing a member
class ErrorWhileRemovingMember(resource.ErrorWhileRemovingMember):
    def __init__(self, group: Group, user: Account):
        super().__init__(group, user)



# APIRouter creates path operations for user module
from typing import Annotated, Literal
from fastapi import APIRouter, Body, Depends, Query, HTTPException

from unipoll_api import dependencies as Dependencies
from unipoll_api.documents import Poll
from unipoll_api.exceptions.resource import APIException
from unipoll_api.actions import PollActions
from unipoll_api.schemas import PollSchemas, QuestionSchemas, PolicySchemas
from unipoll_api import actions

open_router: APIRouter = APIRouter()
router: APIRouter = APIRouter()


query_params = list[Literal["all", "questions", "policies"]]


# Get poll by id
@router.get("/{poll_id}",
            response_description="Poll details",
            response_model=PollSchemas.PollResponse,
            response_model_exclude_none=True)
async def get_poll(poll: Poll = Depends(Dependencies.get_poll),
                   include: Annotated[query_params | None, Query()] = None):
    try:
        params = {}
        if include:
            if "all" in include:
                params = {"include_questions": True, "include_policies": True}
            else:
                if "questions" in include:
                    params = {"include_questions": True}
                if "policies" in include:
                    params = {"include_policies": True}
        return await PollActions.get_poll(poll, **params)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Update poll details
@router.patch("/{poll_id}",
              response_description="Update Poll detail",
              response_model=PollSchemas.PollResponse,
              response_model_exclude_none=True)
async def update_poll(poll: Poll = Depends(Dependencies.get_poll),
                      data: PollSchemas.UpdatePollRequest = Body(...)):
    try:
        return await PollActions.update_poll(poll, data)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Delete poll by id
@router.delete("/{poll_id}",
               response_description="Result of delete operation",
               status_code=204)
async def delete_poll(poll: Poll = Depends(Dependencies.get_poll)):
    try:
        return await PollActions.delete_poll(poll)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Get list of questions in a poll
@router.get("/{poll_id}/questions",
            response_description="Questions in a poll",
            response_model=QuestionSchemas.QuestionList,
            response_model_exclude_none=True)
async def get_questions(poll: Poll = Depends(Dependencies.get_poll),
                        include: Annotated[query_params | None, Query()] = None):
    try:
        return await PollActions.get_poll_questions(poll)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


@router.get("/{poll_id}/policies",
            response_description="Policy list of a poll",
            response_model=PolicySchemas.PolicyList,
            response_model_exclude_none=True)
async def get_policies(poll: Poll = Depends(Dependencies.get_poll),
                       include: Annotated[query_params | None, Query()] = None):
    try:
        return await actions.PolicyActions.get_policies(resource=poll)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))



# FastAPI
from typing import Annotated, Literal
from fastapi import APIRouter, Body, Depends, HTTPException, Query, status
from unipoll_api import dependencies as Dependencies
from unipoll_api import actions
from unipoll_api.exceptions.resource import APIException
from unipoll_api.documents import Account, Workspace, ResourceID, Policy
from unipoll_api.schemas import WorkspaceSchemas, PolicySchemas, GroupSchemas, MemberSchemas, PollSchemas
from unipoll_api import AccountManager

# APIRouter creates path operations for user module
open_router: APIRouter = APIRouter()
# router: APIRouter = APIRouter(dependencies=[Depends(Dependencies.check_workspace_permission)])
router: APIRouter = APIRouter()


# TODO: Move to open router to a separate file
# Get all workspaces with user as a member or owner
@open_router.get("",
                 response_description="List of all workspaces",
                 response_model=WorkspaceSchemas.WorkspaceList)
async def get_workspaces():
    """
    Returns all workspaces where the current user is a member.
    The request does not accept any query parameters.
    """
    try:
        return await actions.WorkspaceActions.get_workspaces()
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Create a new workspace for current user
@open_router.post("",
                  response_description="Created workspaces",
                  status_code=201,
                  response_model=WorkspaceSchemas.WorkspaceCreateOutput)
async def create_workspace(input_data: WorkspaceSchemas.WorkspaceCreateInput = Body(...)):
    """
    Creates a new workspace for the current user.
    Body parameters:
    - **name** (str): name of the workspace, must be unique
    - **description** (str): description of the workspace

    Returns the created workspace information.
    """
    try:
        return await actions.WorkspaceActions.create_workspace(input_data=input_data)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


query_params = list[Literal["all", "policies", "groups", "members", "polls"]]


# Get a workspace with the given id
@router.get("/{workspace_id}",
            response_description="Workspace data",
            response_model=WorkspaceSchemas.Workspace,
            response_model_exclude_defaults=True,
            response_model_exclude_none=True)
async def get_workspace(workspace: Workspace = Depends(Dependencies.get_workspace),
                        include: Annotated[query_params | None, Query()] = None):
    """
    ### Description:
    Endpoint to get a workspace with the given id.
    By default, it returns the basic information of the workspace such as id, name, and description.
    The user can specify other resources to include in the response using the query parameters.

    For example, to include groups and members in the response, the user can send the following GET request:
    > `/workspaces/6497fdbafe12e8ff9017f253?include=groups&include=members`

    To include all resources, the user can send the following GET request:
    > `/workspaces/6497fdbafe12e8ff9017f253?include=all`

    To get basic information of the workspace, the user can send the following GET request:
    > `/workspaces/6497fdbafe12e8ff9017f253`

    ### Path parameters:
    - **workspace_id** (str): id of the workspace

    ### Query parameters:
    - **include** (str): resources to include in the response

        #### Possible values:
        - **groups**: include groups in the response
        - **members**: include members in the response
        - **policies**: include policies in the response
        - **all**: include all resources in the response

    ### Response:
    Returns a workspace with the given id.
    """
    try:
        params = {}
        if include:
            if "all" in include:
                params = {"include_groups": True,
                          "include_members": True,
                          "include_policies": True,
                          "include_polls": True}
            else:
                if "groups" in include:
                    params["include_groups"] = True
                if "members" in include:
                    params["include_members"] = True
                if "policies" in include:
                    params["include_policies"] = True
                if "polls" in include:
                    params["include_polls"] = True
        return await actions.WorkspaceActions.get_workspace(workspace, **params)

    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Update a workspace with the given id
@router.patch("/{workspace_id}", response_description="Updated workspace", response_model=WorkspaceSchemas.Workspace)
async def update_workspace(workspace: Workspace = Depends(Dependencies.get_workspace),
                           input_data: WorkspaceSchemas.WorkspaceUpdateRequest = Body(...)
                           ):
    """
    Updates the workspace with the given id.
    Query parameters:
        @param workspace_id: id of the workspace to update
    Body parameters:
    - **name** (str): name of the workspace, must be unique
    - **description** (str): description of the workspace

    Returns the updated workspace.
    """
    try:
        return await actions.WorkspaceActions.update_workspace(workspace, input_data)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Delete a workspace with the given id
@router.delete("/{workspace_id}",
               response_description="Deleted workspace",
               status_code=204)
async def delete_workspace(workspace: Workspace = Depends(Dependencies.get_workspace)):
    """
    Deletes the workspace with the given id.
    Query parameters:
        @param workspace_id: id of the workspace to delete

    Returns status code 204 if the workspace is deleted successfully.
    Response has no detail.
    """
    try:
        await actions.WorkspaceActions.delete_workspace(workspace)
        return status.HTTP_204_NO_CONTENT
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List all groups in the workspace
@router.get("/{workspace_id}/groups",
            response_description="List of all groups",
            response_model=GroupSchemas.GroupList)
async def get_groups(workspace: Workspace = Depends(Dependencies.get_workspace)):
    try:
        return await actions.GroupActions.get_groups(workspace)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List all groups in the workspace
@router.post("/{workspace_id}/groups",
             status_code=201,
             response_description="Created Group",
             response_model=GroupSchemas.GroupCreateOutput)
async def create_group(workspace: Workspace = Depends(Dependencies.get_workspace),
                       input_data: GroupSchemas.GroupCreateInput = Body(...)):
    try:
        return await actions.GroupActions.create_group(workspace, input_data.name, input_data.description)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List all members in the workspace
@router.get("/{workspace_id}/members",
            response_description="List of all groups",
            response_model=MemberSchemas.MemberList,
            response_model_exclude_unset=True)
async def get_workspace_members(workspace: Workspace = Depends(Dependencies.get_workspace)):
    try:
        return await actions.MembersActions.get_members(workspace)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Add members to the workspace
@router.post("/{workspace_id}/members",
             response_description="List added members",
             response_model=MemberSchemas.MemberList)
async def add_workspace_members(workspace: Workspace = Depends(Dependencies.get_workspace),
                                member_data: MemberSchemas.AddMembers = Body(...)):
    try:
        return await actions.MembersActions.add_members(workspace, member_data.accounts)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Remove member from the workspace
@router.delete("/{workspace_id}/members/{account_id}",
               response_description="Updated list removed members",
               response_model_exclude_unset=True)
async def remove_workspace_member(workspace: Workspace = Depends(Dependencies.get_workspace),
                                  account: Account = Depends(Dependencies.get_account)):
    try:
        return await actions.MembersActions.remove_member(workspace, account)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List all policies in the workspace
@router.get("/{workspace_id}/policies",
            response_description="List of all policies",
            response_model=PolicySchemas.PolicyList)
async def get_workspace_policies(workspace: Workspace = Depends(Dependencies.get_workspace)):
    try:
        return await actions.PolicyActions.get_policies(resource=workspace)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# List member's permissions in the workspace
@router.get("/{workspace_id}/policy",
            response_description="List member policy(permissions)",
            response_model=PolicySchemas.PolicyOutput)
async def get_workspace_policy(workspace: Workspace = Depends(Dependencies.get_workspace),
                               account_id: ResourceID | None = None):
    try:
        account = await Dependencies.get_account(account_id) if account_id else AccountManager.active_user.get()
        policy_list = await actions.PolicyActions.get_policies(resource=workspace, policy_holder=account)
        policy = policy_list.policies[0]
        return PolicySchemas.PolicyOutput(**policy.model_dump())
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Set permissions for a member in a workspace
@router.put("/{workspace_id}/policy",
            response_description="Updated permissions",
            response_model=PolicySchemas.PolicyOutput)
async def set_workspace_policy(workspace: Workspace = Depends(Dependencies.get_workspace),
                               permissions: PolicySchemas.PolicyInput = Body(...)):
    """
    Sets the permissions for a user in a workspace.
    Query parameters:
        @param workspace_id: id of the workspace to update
    Body parameters:
    - **user_id** (str): id of the user to update
    - **permissions** (int): new permissions for the user

    Returns the updated workspace.
    """
    try:
        # return await WorkspaceActions.set_workspace_policy(workspace, permissions)
        policy = None
        if permissions.policy_id:
            policy = await Dependencies.get_policy(permissions.policy_id)  # type: ignore
        elif permissions.account_id:
            account = await Dependencies.get_account(permissions.account_id)
            # policy = await Policy.find_one(Policy.policy_holder.id == account.id, fetch_links=True)
            # Temporarily workaround
            policy_list = await actions.PolicyActions.get_policies(resource=workspace, policy_holder=account)
            policy = policy_list.policies[0]  # type: ignore
            policy = await Policy.get(policy.id, fetch_links=True)  # type: ignore
        elif permissions.group_id:
            # Temporarily workaround
            group = await Dependencies.get_group(permissions.group_id)
            policy_list = await actions.PolicyActions.get_policies(resource=workspace, policy_holder=group)
            policy = policy_list.policies[0]  # type: ignore
            policy = await Policy.get(policy.id, fetch_links=True)  # type: ignore
        else:
            account = AccountManager.active_user.get()
            policy_list = await actions.PolicyActions.get_policies(resource=workspace, policy_holder=account)
            policy = policy_list.policies[0]  # type: ignore
            policy = await Policy.get(policy.id, fetch_links=True)  # type: ignore

        if not policy:
            raise APIException(404, "Policy not found 404")

        return await actions.PolicyActions.update_policy(policy, new_permissions=permissions.permissions)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Get All Workspace Permissions
@open_router.get("/permissions",
                 response_description="List of all workspace permissions",
                 response_model=PolicySchemas.PermissionList)
async def get_workspace_permissions():
    try:
        return await actions.PermissionsActions.get_workspace_permissions()
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Get Workspace Polls
@router.get("/{workspace_id}/polls",
            response_description="List of all polls in the workspace",
            response_model=PollSchemas.PollList,
            response_model_exclude_none=True)
async def get_polls(workspace: Workspace = Depends(Dependencies.get_workspace)):
    try:
        return await actions.PollActions.get_polls(workspace)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Create a new poll in the workspace
@router.post("/{workspace_id}/polls",
             response_description="Created poll",
             status_code=201,
             response_model=PollSchemas.PollResponse)
async def create_poll(workspace: Workspace = Depends(Dependencies.get_workspace),
                      input_data: PollSchemas.CreatePollRequest = Body(...)):
    try:
        return await actions.PollActions.create_poll(workspace, input_data)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))



from contextvars import ContextVar
from typing import Optional, AsyncGenerator
from beanie import PydanticObjectId
from fastapi import Depends, Request
from fastapi_users import BaseUserManager, FastAPIUsers
from fastapi_users.authentication import AuthenticationBackend, CookieTransport
from unipoll_api.utils.auth_transport import BearerTransport
from unipoll_api.config import get_settings
from fastapi_users_db_beanie import BeanieUserDatabase, ObjectIDIDMixin
from unipoll_api.utils.token_db import BeanieAccessTokenDatabase
from unipoll_api.utils.auth_strategy import DatabaseStrategy
from unipoll_api.documents import Account, AccessToken
from unipoll_api.utils import colored_dbg


SECRET = get_settings().secrete_key

UserManager = BaseUserManager[Account, PydanticObjectId]  # type: ignore


class AccountManager(ObjectIDIDMixin, UserManager):
    reset_password_token_secret = SECRET
    verification_token_secret = SECRET

    async def on_after_register(self, user: Account,
                                request: Optional[Request] = None) -> None:
        # print(f"Account {user.id} has registered.")
        colored_dbg.info(f"Account {user.id} has registered.")

    async def on_after_forgot_password(self, user: Account, token: str,
                                       request: Optional[Request] = None
                                       ) -> None:
        colored_dbg.info(
            f"Account {user.id} has forgot their password. Reset token: {token}")

    async def on_after_request_verify(self, user: Account, token: str,
                                      request: Optional[Request] = None
                                      ) -> None:
        colored_dbg.info(
            "Verification requested for user {}. \
            Verification token: {}".format(user.id, token))

    async def on_before_delete(self, user: Account,
                               request: Optional[Request] = None) -> None:
        colored_dbg.info(
            f"Account {user.id} is going to be deleted, cleaning up their data.")


async def get_account_db() -> AsyncGenerator:
    yield BeanieUserDatabase(Account)  # type: ignore


async def get_user_manager(user_db=Depends(get_account_db)) -> AsyncGenerator[UserManager, None]:
    yield AccountManager(user_db)


bearer_transport = BearerTransport(tokenUrl="auth/jwt/login")
cookie_transport = CookieTransport(cookie_max_age=3600)


async def get_access_token_db():
    yield BeanieAccessTokenDatabase(AccessToken)  # type: ignore


def get_database_strategy(access_token_db=Depends(get_access_token_db)) -> DatabaseStrategy:
    return DatabaseStrategy(access_token_db, lifetime_seconds=3600)


jwt_backend = AuthenticationBackend(
    name="token",
    transport=bearer_transport,
    get_strategy=get_database_strategy,
)
cookie_backend = AuthenticationBackend(
    name="cookie",
    transport=cookie_transport,
    get_strategy=get_database_strategy,
)

fastapi_users = FastAPIUsers[Account, PydanticObjectId](get_user_manager, [jwt_backend, cookie_backend])  # type: ignore

get_current_active_user = fastapi_users.current_user(active=True)
active_user: ContextVar = ContextVar("active_user")



from typing import Annotated
from fastapi import APIRouter, Body, Depends, HTTPException, Header, status
from fastapi.security import OAuth2PasswordRequestForm
from fastapi_users import BaseUserManager, models
from fastapi_users.openapi import OpenAPIResponseType
from fastapi_users.router.common import ErrorCode, ErrorModel
from fastapi_users.authentication import Strategy

from unipoll_api import account_manager as AccountManager

# import fastapi_users, get_user_manager, jwt_backend, get_database_strategy, get_access_token_db
from unipoll_api.actions import authentication as AuthActions
# from unipoll_api.schemas import authentication as AuthSchemas
from unipoll_api.schemas import account as AccountSchemas
from unipoll_api.exceptions.resource import APIException
from unipoll_api.utils.token_db import BeanieAccessTokenDatabase
router: APIRouter = APIRouter()


login_responses: OpenAPIResponseType = {
    status.HTTP_400_BAD_REQUEST: {
        "model": ErrorModel,
        "content": {
            "application/json": {
                "examples": {
                    ErrorCode.LOGIN_BAD_CREDENTIALS: {
                        "summary": "Bad credentials or the user is inactive.",
                        "value": {"detail": ErrorCode.LOGIN_BAD_CREDENTIALS},
                    },
                    ErrorCode.LOGIN_USER_NOT_VERIFIED: {
                        "summary": "The user is not verified.",
                        "value": {"detail": ErrorCode.LOGIN_USER_NOT_VERIFIED},
                    },
                }
            }
        },
    },
    **AccountManager.jwt_backend.transport.get_openapi_login_responses_success(),
}


@router.post(
    "/jwt/login",
    name=f"auth:{AccountManager.jwt_backend.name}.login",
    responses=login_responses,
    response_model_exclude_unset=True
)
async def login(credentials: OAuth2PasswordRequestForm = Depends(),
                user_manager: BaseUserManager[models.UP, models.ID] = Depends(AccountManager.get_user_manager),
                token_db: BeanieAccessTokenDatabase = Depends(AccountManager.get_access_token_db),
                strategy: Strategy = Depends(AccountManager.get_database_strategy)):
    user = await user_manager.authenticate(credentials)

    if user is None or not user.is_active:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=ErrorCode.LOGIN_BAD_CREDENTIALS,
        )
    # if requires_verification and not user.is_verified:
    #     raise HTTPException(
    #         status_code=status.HTTP_400_BAD_REQUEST,
    #         detail=ErrorCode.LOGIN_USER_NOT_VERIFIED,
    #     )

    return await AccountManager.jwt_backend.login(strategy, user)


# Refresh the access token using the refresh token
@router.post("/jwt/refresh", responses=login_responses, response_model_exclude_unset=True)
async def refresh_jwt(authorization: Annotated[str, Header(...)],
                      refresh_token: Annotated[str, Header(...)],
                      token_db: BeanieAccessTokenDatabase = Depends(AccountManager.get_access_token_db),
                      strategy: Strategy = Depends(AccountManager.get_database_strategy)):
    """Refresh the access token using the refresh token.

    Headers:
        authorization: `Authorization` header with the access token
        refresh_token: `Refresh-Token` header with the refresh token
    """
    try:
        return await AuthActions.refresh_token(authorization, refresh_token)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=e.detail)


# Refresh the access token using the refresh token and Client ID
@router.post("/jwt/postman_refresh", responses=login_responses, response_model_exclude_unset=True)
async def refresh_jwt_with_client_ID(authorization: Annotated[str, Header(...)],
                                     body: Annotated[str, Body(...)],
                                     token_db: BeanieAccessTokenDatabase = Depends(AccountManager.get_access_token_db),
                                     strategy: Strategy = Depends(AccountManager.get_database_strategy)):
    """Refresh the access token using the refresh token.

    Headers:
        authorization: `Authorization` header with the access token
    Body:
        refresh_token: `Refresh-Token` header with the refresh token
    """
    try:
        # import json
        # print(body.decode('utf-8'))
        # body = json.loads(body.decode('utf-8'))
        # print(body)
        # AuthSchemas.PostmanRefreshTokenRequest(**body)
        return await AuthActions.refresh_token_with_clientID(authorization, body, token_db, strategy)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=e.detail)


# Include prebuilt routes for authentication
router.include_router(AccountManager.fastapi_users.get_register_router(
    AccountSchemas.Account, AccountSchemas.CreateAccount))
router.include_router(AccountManager.fastapi_users.get_reset_password_router())
router.include_router(AccountManager.fastapi_users.get_verify_router(AccountSchemas.Account))



from pydantic import ConfigDict, BaseModel, Field
from typing import Optional
from unipoll_api.documents import ResourceID


# Schema for the response with basic workspace info (name and role)
class Workspace(BaseModel):
    id: Optional[ResourceID] = None
    name: Optional[str] = None
    description: Optional[str] = None
    members: Optional[list] = None
    groups: Optional[list] = None
    policies: Optional[list] = None
    polls: Optional[list] = None
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "id": "1a2b3c4d5e6f7g8h9i0j",
            "name": "Workspace 01",
            "description": "This is an example workspace",
        }
    })


class WorkspaceShort(BaseModel):
    id: ResourceID = Field(title="ID")
    name: str = Field(title="Name")
    description: str = Field(title="Description")


# Schema for the response with a list of workspaces
# It can be used to return a list of workspaces with basic info or full info
class WorkspaceList(BaseModel):
    workspaces: list[WorkspaceShort] | list[Workspace]
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "workspaces": [
                {
                    "name": "Workspace 01",
                    "description": "This is an example workspace",
                    "owner": "true",
                },
                {
                    "name": "Workspace 02",
                    "description": "This is another example workspace",
                    "owner": "false",
                },
            ]
        }
    })


# Schema for the request to create a workspace
class WorkspaceCreateInput(BaseModel):
    name: str = Field(title="Name")
    description: str = Field(title="Description")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Workspace 01",
            "description": "This is an example workspace",
        }
    })


# Schema for the request to update a workspace
class WorkspaceUpdateRequest(BaseModel):
    name: Optional[str] = Field(None, title="Name")
    description: Optional[str] = Field(None, title="Description")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Workspace 01",
            "description": "This is an example workspace",
        }
    })


# Schema for the response when a workspace is created
class WorkspaceCreateOutput(BaseModel):
    id: ResourceID = Field(title="ID")
    name: str = Field(title="Name")
    description: str = Field(title="Description")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Workspace 01",
            "description": "This is an example workspace",
        }
    })



from unipoll_api import AccountManager
from unipoll_api.documents import Account, Workspace, Group, Policy, Resource
from unipoll_api.schemas import MemberSchemas, PolicySchemas
from unipoll_api.exceptions import PolicyExceptions, ResourceExceptions
from unipoll_api.utils import Permissions
from unipoll_api.utils.permissions import check_permissions


# Helper function to get policies from a resource
# NOTE: This can be moved to utils
async def get_policies_from_resource(resource: Resource) -> list[Policy]:
    policies: list[Policy] = []
    try:
        await check_permissions(resource, "get_policies")
        return resource.policies  # type: ignore
    except ResourceExceptions.UserNotAuthorized:
        account = AccountManager.active_user.get()
        for policy in resource.policies:
            if policy.policy_holder.ref.id == account.id:  # type: ignore
                policies.append(policy)  # type: ignore
        return policies


# Get all policies of a workspace
async def get_policies(policy_holder: Account | Group | None = None,
                       resource: Resource | None = None) -> PolicySchemas.PolicyList:
    policy_list = []
    policy: Policy
    all_policies = []

    # Get policies from a specific resource
    if resource:
        all_policies = await get_policies_from_resource(resource)
    # Get policies from all resources
    else:
        all_workspaces = Workspace.find(fetch_links=True)
        all_groups = Group.find(fetch_links=True)
        all_resources = await all_workspaces.to_list() + await all_groups.to_list()

        for resource in all_resources:
            all_policies += await get_policies_from_resource(resource)

    # Build policy list
    for policy in all_policies:
        # Filter by policy_holder if specified
        if policy_holder:
            if (policy.policy_holder.ref.id != policy_holder.id):
                continue
        policy_list.append(await get_policy(policy, False))
    # Return policy list
    return PolicySchemas.PolicyList(policies=policy_list)


# @check_permissions(resource=policy.parent_resource, required_permissions="get_policy", permission_check=True)
async def get_policy(policy: Policy, permission_check: bool = True) -> PolicySchemas.PolicyShort:
    await policy.parent_resource.fetch_all_links()  # type: ignore
    await check_permissions(policy.parent_resource, "get_policies", permission_check)

    # Convert policy_holder link to Member object
    ph_type = policy.policy_holder_type
    ph_ref = policy.policy_holder.ref.id
    policy_holder = await Account.get(ph_ref) if ph_type == "account" else await Group.get(ph_ref)
    if not policy_holder:
        raise PolicyExceptions.PolicyHolderNotFound(ph_ref)
    policy_holder = MemberSchemas.Member(**policy_holder.model_dump())  # type: ignore
    resource_type: str = policy.parent_resource.resource_type  # type: ignore
    PermissionType = eval("Permissions." + resource_type.capitalize() + "Permissions")
    permissions = PermissionType(policy.permissions).name.split('|')  # type: ignore
    return PolicySchemas.PolicyShort(id=policy.id,
                                     policy_holder_type=policy.policy_holder_type,
                                     policy_holder=policy_holder.model_dump(exclude_unset=True),
                                     permissions=permissions)


async def update_policy(policy: Policy,
                        new_permissions: list[str],
                        check_permissions: bool = True) -> PolicySchemas.PolicyOutput:

    # BUG: since the parent_resource is of multiple types, it is not fetched properly, so we fetch it manually
    await policy.parent_resource.fetch_all_links()  # type: ignore

    # Check if the user has the required permissions to update the policy
    await Permissions.check_permissions(policy.parent_resource, "update_policies", check_permissions)
    ResourcePermissions = eval(
        "Permissions." + policy.parent_resource.resource_type.capitalize() + "Permissions")  # type: ignore

    # Calculate the new permission value from request
    new_permission_value = 0
    for i in new_permissions:
        try:
            new_permission_value += ResourcePermissions[i].value  # type: ignore
        except KeyError:
            raise ResourceExceptions.InvalidPermission(i)
    # Update permissions
    policy.permissions = ResourcePermissions(new_permission_value)  # type: ignore
    await Policy.save(policy)

    if policy.policy_holder_type == "account":
        policy_holder = await Account.get(policy.policy_holder.ref.id)
    elif policy.policy_holder_type == "group":
        policy_holder = await Group.get(policy.policy_holder.ref.id)

    return PolicySchemas.PolicyOutput(
        permissions=ResourcePermissions(policy.permissions).name.split('|'),  # type: ignore
        policy_holder=policy_holder.model_dump())  # type: ignore



from fastapi import status
from unipoll_api.exceptions import resource
# from app.models.documents import ResourceID


class InvalidAuthorizationHeader(resource.APIException):
    def __init__(self):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail="Invalid Authorization header")


class InvalidAccessToken(resource.APIException):
    def __init__(self):
        super().__init__(code=status.HTTP_401_UNAUTHORIZED,
                         detail="Invalid Access token")


class InvalidRefreshToken(resource.APIException):
    def __init__(self):
        super().__init__(code=status.HTTP_401_UNAUTHORIZED,
                         detail="Invalid Refresh token")


class refreshTokenExpired(resource.APIException):
    def __init__(self):
        super().__init__(code=status.HTTP_401_UNAUTHORIZED,
                         detail="Refresh token expired; Please login again")


class InvalidClientID(resource.APIException):
    def __init__(self):
        super().__init__(code=status.HTTP_401_UNAUTHORIZED,
                         detail="Invalid Client ID")



from fastapi import APIRouter, Depends
from unipoll_api.dependencies import set_active_user

# Impport endpoints defined in the routes directory
from . import account as AccountRoutes
from . import authentication as AuthenticationRoutes
from . import group as GroupRoutes
from . import poll as PollRoutes
from . import websocket as WebSocketRoutes
from . import workspace as WorkspaceRoutes

# Create main router
router: APIRouter = APIRouter()

# Add endpoints defined in the routes directory
router.include_router(WorkspaceRoutes.open_router,
                      prefix="/workspaces",
                      tags=["Workspaces"],
                      dependencies=[Depends(set_active_user)])
router.include_router(WorkspaceRoutes.router,
                      prefix="/workspaces",
                      tags=["Workspaces"],
                      dependencies=[Depends(set_active_user)])
router.include_router(GroupRoutes.open_router,
                      prefix="/groups",
                      tags=["Groups"],
                      dependencies=[Depends(set_active_user)])
router.include_router(GroupRoutes.router,
                      prefix="/groups",
                      tags=["Groups"],
                      dependencies=[Depends(set_active_user)])
router.include_router(PollRoutes.router,
                      prefix="/polls",
                      tags=["Polls"],
                      dependencies=[Depends(set_active_user)])
router.include_router(WebSocketRoutes.router,
                      prefix="/ws",
                      tags=["WebSocket"])
router.include_router(AccountRoutes.router,
                      prefix="/accounts",
                      tags=["Accounts"],
                      dependencies=[Depends(set_active_user)])
router.include_router(AuthenticationRoutes.router,
                      prefix="/auth",
                      tags=["Authentication"])
router.include_router(WebSocketRoutes.router,
                      prefix="/ws",
                      tags=["WebSocket"])



from pydantic import ConfigDict, BaseModel, Field
from typing import Any, Optional
from unipoll_api.documents import ResourceID


class Group(BaseModel):
    id: Optional[ResourceID] = None
    name: Optional[str] = None
    description: Optional[str] = None
    workspace: Optional[Any] = None
    # groups: Optional[list]
    members: Optional[list] = None
    policies: Optional[list] = None


# Schema for the response with basic group info after creation
class GroupShort(BaseModel):
    id: ResourceID
    name: str
    description: str


# Schema for the response with basic group info
class GroupList(BaseModel):
    groups: list[GroupShort] | list[Group]


# Schema for the request to create a new group
class GroupCreateRequest(BaseModel):
    name: str = Field(default="", min_length=3, max_length=50)
    workspace: ResourceID = Field(title="Workspace ID")
    description: str = Field(default="", title="Description", max_length=300)
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Group 01",
            "workspace": "60b9d1c8e1f1d5f5f5b4f8e1",
            "description": "My first Group",
        }
    })


# Schema for the request to create a new group
class GroupCreateInput(BaseModel):
    name: str = Field(default="", min_length=3, max_length=50)
    description: str = Field(default="", title="Description", max_length=300)
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "Group 01",
            "description": "My first Group",
        }
    })


# Schema for the response to a group creation request
class GroupCreateOutput(BaseModel):
    id: ResourceID
    name: str
    description: str


# Schema for the request to add a user to a group
class GroupUpdateRequest(BaseModel):
    name: Optional[str] = Field(None, title="Name", min_length=3, max_length=50)
    description: Optional[str] = Field(default="", title="Description", max_length=300)
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "Description": "Updated description"
        }
    })



import re
import base64
from beanie import PydanticObjectId

from fastapi import Depends

from unipoll_api.documents import Account
from unipoll_api.exceptions import AuthExceptions, AccountExceptions
from unipoll_api.utils import colored_dbg as Debug
from unipoll_api.account_manager import jwt_backend, get_database_strategy, get_access_token_db


async def refresh_token(authorization: str,
                        refresh_token: str,
                        token_db=Depends(get_access_token_db),
                        strategy=Depends(get_database_strategy)):
    # Make sure the Authorization header is valid and extract the access token
    try:
        access_token = re.match(r'^Bearer ([A-z0-9\-]+)$', authorization).group(1)  # type: ignore
    except Exception as e:
        Debug.print_error(str(e))
        raise AuthExceptions.InvalidAuthorizationHeader()

    # Get the token data from the database using the access token
    # NOTE: We do not supply a max_age parameter in case access tokens has already expired
    token_data = await token_db.get_by_token(access_token)

    # Make sure the access token exists in the database
    if token_data is None:
        raise AuthExceptions.InvalidAccessToken()

    # Make sure the access token is associated with the supplied refresh token
    if token_data.refresh_token != refresh_token:
        raise AuthExceptions.InvalidRefreshToken()

    # Get the user from the database using the user ID in the token data
    user = await Account.get(token_data.user_id)
    if user is None:
        raise AccountExceptions.AccountNotFound(token_data.user_id)

    # Check if the refresh token is the most recent one
    all_tokens = await token_db.get_token_family_by_user_id(user.id)
    if (await all_tokens.to_list())[0].refresh_token != refresh_token:
        # If not, delete all tokens associated with the user and return an error
        await strategy.destroy_token_family(user)
        raise AuthExceptions.refreshTokenExpired()

    # Login the user using the supplied strategy
    # Generate new pair of access and refresh tokens
    # Returns Response object with LoginResponse schema
    return await jwt_backend.login(strategy, user)


async def refresh_token_with_clientID(authorization: str,
                                      body: str,
                                      token_db=Depends(get_access_token_db),
                                      strategy=Depends(get_database_strategy)):
    # Make sure the Authorization header is valid and extract the access token
    try:
        client_id = re.match(r'^Basic (\S+)$', authorization).group(1)  # type: ignore
        refresh_token = re.match(r'^refresh_token=(\S+)&grant_type=refresh_token$', body).group(1)  # type: ignore
    except Exception as e:
        Debug.print_error(str(e))
        raise AuthExceptions.InvalidAuthorizationHeader()

    # Get the token data from the database using the access token
    token_data = await token_db.get_by_refresh_token(refresh_token)

    # Make sure the access token exists in the database
    if token_data is None:
        raise AuthExceptions.InvalidAccessToken()

    # Get the user from the database using the user ID in the token data
    user = await Account.get(token_data.user_id)
    if user is None:
        raise AccountExceptions.AccountNotFound(token_data.user_id)

    # Decode the client ID and make sure it matches account ID
    client_id = base64.b64decode(client_id)
    if PydanticObjectId(str(client_id, "utf-8")[:-1]) != user.id:
        raise AuthExceptions.InvalidClientID()

    # Check if the refresh token is the most recent one
    all_tokens = await token_db.get_token_family_by_user_id(user.id)
    if (await all_tokens.to_list())[0].refresh_token != refresh_token:
        # If not, delete all tokens associated with the user and return an error
        await strategy.destroy_token_family(user)
        raise AuthExceptions.refreshTokenExpired()

    # Login the user using the supplied strategy
    # Generate new pair of access and refresh tokens
    # Returns Response object with LoginResponse schema
    return await jwt_backend.login(strategy, user)



# from typing import ForwardRef, NewType, TypeAlias, Optional
from typing import Literal
from bson import DBRef
from beanie import BackLink, Document, WriteRules, after_event, Insert, Link, PydanticObjectId  # BackLink
from fastapi_users_db_beanie import BeanieBaseUser
from pydantic import Field
from unipoll_api.utils import colored_dbg as Debug
from unipoll_api.utils.token_db import BeanieBaseAccessToken


# Create a link to the Document model
async def create_link(document: Document) -> Link:
    ref = DBRef(collection=document._document_settings.name,  # type: ignore
                id=document.id)
    link = Link(ref, type(document))
    return link


# Custom PydanticObjectId class to override due to a bug
class ResourceID(PydanticObjectId):
    @classmethod
    def __modify_schema__(cls, field_schema):  # type: ignore
        field_schema.update(
            type="string",
            example="5eb7cf5a86d9755df3a6c593",
        )


class AccessToken(BeanieBaseAccessToken, Document):  # type: ignore
    pass


class Resource(Document):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    resource_type: Literal["workspace", "group", "poll"]
    name: str = Field(title="Name", description="Name of the resource", min_length=3, max_length=50)
    description: str = Field(default="", title="Description", max_length=1000)
    policies: list[Link["Policy"]] = []

    @after_event(Insert)
    def create_group(self) -> None:
        Debug.info(f'New {self.resource_type} "{self.id}" has been created')

    async def add_policy(self, member: "Group | Account", permissions, save: bool = True) -> None:
        new_policy = Policy(policy_holder_type='account',
                            policy_holder=(await create_link(member)),
                            permissions=permissions,
                            parent_resource=self)  # type: ignore

        # Add the policy to the group
        self.policies.append(new_policy)  # type: ignore
        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore

    async def remove_policy(self, member: "Group | Account", save: bool = True) -> None:
        for policy in self.policies:
            if policy.policy_holder.id == member.id:  # type: ignore
                self.policies.remove(policy)
                if save:
                    await self.save(link_rule=WriteRules.WRITE)  # type: ignore


class Account(BeanieBaseUser, Document):  # type: ignore
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    first_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    last_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")


class Workspace(Resource):
    resource_type: Literal["workspace"] = "workspace"
    members: list[Link["Account"]] = []
    groups: list[Link["Group"]] = []
    polls: list[Link["Poll"]] = []

    async def add_member(self, account: "Account", permissions, save: bool = True) -> "Account":
        # Add the account to the group
        self.members.append(account)  # type: ignore
        # Create a policy for the new member
        await self.add_policy(account, permissions, save=False)  # type: ignore
        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return account

    async def remove_member(self, account, save: bool = True) -> bool:
        # Remove the account from the group
        for i, member in enumerate(self.members):
            if account.id == member.id:  # type: ignore
                self.members.remove(member)
                Debug.info(f"Removed member {member.id} from {self.resource_type} {self.id}")  # type: ignore
                break

        # Remove the policy from the workspace
        await self.remove_policy(account, save=False)  # type: ignore

        # Remove the member from all groups in the workspace
        group: Group
        for group in self.groups:  # type: ignore
            await group.remove_member(account, save=False)
            await group.remove_policy(account, save=False)
            await Group.save(group, link_rule=WriteRules.WRITE)

        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return True


class Group(Resource):
    resource_type: Literal["group"] = "group"
    workspace: BackLink[Workspace] = Field(original_field="groups")
    members: list[Link["Account"]] = []
    groups: list[Link["Group"]] = []

    async def add_member(self, account: "Account", permissions, save: bool = True) -> "Account":
        if account not in self.workspace.members:  # type: ignore
            from unipoll_api.exceptions import WorkspaceExceptions
            raise WorkspaceExceptions.UserNotMember(self.workspace, account)  # type: ignore

        # Add the account to the group
        self.members.append(account)  # type: ignore
        # Create a policy for the new member
        await self.add_policy(account, permissions, save=False)  # type: ignore
        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return account

    async def remove_member(self, account, save: bool = True) -> bool:
        # Remove the account from the group
        for i, member in enumerate(self.members):
            if account.id == member.id:  # type: ignore
                self.members.remove(member)
                Debug.info(f"Removed member {member.id} from {self.resource_type} {self.id}")  # type: ignore
                break

        # Remove the policy from the group
        await self.remove_policy(account, save=False)  # type: ignore

        if save:
            await self.save(link_rule=WriteRules.WRITE)  # type: ignore
        return True


class Policy(Document):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    parent_resource: BackLink["Workspace"] | BackLink["Group"] | BackLink["Poll"] = Field(original_field="policies")
    policy_holder_type: Literal["account", "group"]
    policy_holder: Link["Group"] | Link["Account"]
    permissions: int


class Poll(Resource):
    id: ResourceID = Field(default_factory=ResourceID, alias="_id")
    workspace: BackLink["Workspace"] = Field(original_field="polls")
    resource_type: Literal["poll"] = "poll"
    public: bool
    published: bool
    questions: list
    policies: list[Link["Policy"]]


# NOTE: model_rebuild is used to avoid circular imports
Resource.model_rebuild()
Workspace.model_rebuild()
Group.model_rebuild()
Policy.model_rebuild()
Poll.model_rebuild()



from . import account as AccountActions  # noqa: F401
from . import group as GroupActions  # noqa: F401
from . import policy as PolicyActions  # noqa: F401
from . import poll as PollActions  # noqa: F401
from . import authentication as AuthActions  # noqa: F401
from . import workspace as WorkspaceActions  # noqa: F401
from . import permissions as PermissionsActions  # noqa: F401
from . import members as MembersActions  # noqa: F401



from fastapi import APIRouter, status, HTTPException, Depends
from unipoll_api.account_manager import fastapi_users
from unipoll_api.actions import AccountActions
from unipoll_api.exceptions.resource import APIException
from unipoll_api.documents import Account
from unipoll_api.dependencies import get_account
from unipoll_api.schemas import AccountSchemas


# APIRouter creates path operations for user module
router: APIRouter = APIRouter()


@router.get("",
            response_model=AccountSchemas.AccountList)
async def get_all_accounts():
    try:
        accounts = [AccountSchemas.AccountShort(**a.model_dump()) for a in await Account.find_all().to_list()]
        return AccountSchemas.AccountList(accounts=accounts)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Delete current user account
@router.delete("/me",
               status_code=status.HTTP_204_NO_CONTENT)
async def delete_my_account():
    """
        ## Delete current user account

        This route deletes the account of the currently logged in user.

        ### Request body

        - **user** - User object

        ### Expected Response

        **204** - *The account has been deleted*
    """
    try:
        await AccountActions.delete_account()
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Delete user account by id
@router.delete("/{id}",
               status_code=status.HTTP_204_NO_CONTENT)
async def delete_user(account: Account = Depends(get_account)):
    """
        ## Delete current user account

        This route deletes the account of the currently logged in user.

        ### Request body

        - **user** - User object

        ### Expected Response

        **204** - *The account has been deleted*
    """
    try:
        await AccountActions.delete_account(account)
    except APIException as e:
        raise HTTPException(status_code=e.code, detail=str(e))


# Update current user account
router.include_router(fastapi_users.get_users_router(AccountSchemas.Account, AccountSchemas.UpdateAccount))



from . import auth_strategy as AuthStrategy  # noqa: F401
from . import auth_transport as AuthTransport  # noqa: F401
from . import cli_args as ArgParser  # noqa: F401
from . import colored_dbg as Debug  # noqa: F401
from . import path_operations as PathOperations  # noqa: F401
from . import permissions as Permissions  # noqa: F401
from . import token_db as TokenDB  # noqa: F401



from enum import IntFlag
import unipoll_api
from unipoll_api import exceptions

# import functools
# import ast
# from pathlib import Path


# Define the permissions base class as an IntFlag Enum
# The enumerator entries are combination of key: value (permission: #value) pairs,
# where value is an int powers of 2:
# permission1 = 1, 2^0, 0001
# permission2 = 2, 2^1, 0010
# permission3 = 4, 2^2, 0100
# permission4 = 8, 2^3, 1000
Permissions = IntFlag

WorkspacePermissions = IntFlag("WorkspacePermissions", ['get_workspace',
                                                        'update_workspace',
                                                        'delete_workspace',
                                                        'get_members',
                                                        'add_members',
                                                        'remove_members',
                                                        'get_groups',
                                                        'add_groups',
                                                        'update_groups',
                                                        'delete_groups',
                                                        'get_policies',
                                                        'add_policies',
                                                        'update_policies',
                                                        'delete_policies',
                                                        'get_polls',
                                                        'create_polls',
                                                        'delete_polls'])

GroupPermissions = IntFlag("GroupPermissions", ['get_group',
                                                'update_group',
                                                'delete_group',
                                                'get_members',
                                                'add_members',
                                                'remove_members',
                                                'get_policies',
                                                'add_policies',
                                                'update_policies',
                                                'delete_policies'])

PollPermissions = IntFlag("PollPermissions", ['get_poll',
                                              'get_questions',
                                              'update_poll',
                                              'delete_poll',
                                              'get_policies',
                                              'add_policies',
                                              'update_policies',
                                              'delete_policies'
                                              ])

WORKSPACE_ALL_PERMISSIONS = WorkspacePermissions(-1)  # type: ignore
WORKSPACE_BASIC_PERMISSIONS = WorkspacePermissions(sum([WorkspacePermissions["get_workspace"],
                                                       WorkspacePermissions["get_members"],
                                                       WorkspacePermissions["get_polls"]]))  # type: ignore

GROUP_ALL_PERMISSIONS = GroupPermissions(-1)  # type: ignore
GROUP_BASIC_PERMISSIONS = (GroupPermissions["get_group"])  # type: ignore

POLL_ALL_PERMISSIONS = PollPermissions(-1)  # type: ignore
POLL_BASIC_PERMISSIONS = (PollPermissions["get_poll"])  # type: ignore


# Check if a user has a permission
def compare_permissions(user_permission: Permissions, required_permission: Permissions) -> bool:
    """Check if a user has a right provided in the permission argument.
    If the user is not found or has no permission, the default permission NONE is used.
    In which case the function returns False, unless the required permission is also NONE.

    Args:
        :param user_permissions: Dictionary with keys as users and their permissions as the values.
        :required_permission: Required permissions.

    Returns:
        bool: True if the user has the required permission, False otherwise.
    """
    return bool((user_permission & required_permission) == required_permission)


# TODO: Rename
async def get_all_permissions(resource, account) -> Permissions:
    permission_sum = 0
    # print("resource: ", resource.name)
    # await resource.fetch_link("policies")
    # print("policies: ", resource.policies)
    # Get policies for the resource
    for policy in resource.policies:
        # Get policy for the user
        if policy.policy_holder_type == "account":
            policy_holder_id = None
            if hasattr(policy.policy_holder, "id"):     # In case the policy_holder is an Account Document
                policy_holder_id = policy.policy_holder.id
            elif hasattr(policy.policy_holder, "ref"):  # In case the policy_holder is a Link
                policy_holder_id = policy.policy_holder.ref.id
            if policy_holder_id == account.id:
                # print("Found policy for user")
                permission_sum |= policy.permissions
                # print("User permissions: ", policy.permissions)
        # If there is a group that user is a member of, add group permissions to the user permissions
        elif policy.policy_holder_type == "group":
            # Try to fetch the group
            group = await policy.policy_holder.fetch()
            # BUG: sometimes links are not fetched properly
            # If fetching policy holder is not working, find the group manually
            if not group:
                from unipoll_api.documents import Group
                group = await Group.get(policy.policy_holder.ref.id)

            if group:
                await group.fetch_link("policies")
                # print("Checking group: ", group.name)
                if await get_all_permissions(group, account):
                    permission_sum |= policy.permissions
                    # print("Group permissions: ", policy.permissions)

    return permission_sum  # type: ignore


def convert_string_to_permission(resource_type: str, string: str):
    try:
        # return eval(resource_type.capitalize() + "Permissions")[string]
        if resource_type == "workspace":  # type: ignore
            req_permissions = WorkspacePermissions[string]  # type: ignore
        elif resource_type == "group":  # type: ignore
            req_permissions = GroupPermissions[string]  # type: ignore
        elif resource_type == "poll":  # type: ignore
            req_permissions = PollPermissions[string]  # type: ignore
        else:
            raise ValueError("Unknown resource type")
        return req_permissions
    except NameError:
        raise ValueError("Invalid permission string")


async def check_permissions(resource, required_permissions: str | list[str] | None = None, permission_check=True):
    if permission_check and required_permissions:
        account = unipoll_api.AccountManager.active_user.get()           # Get the active user
        user_permissions = await get_all_permissions(resource, account)  # Get the user permissions
        if isinstance(required_permissions, str):  # If only one permission is required
            required_permissions = [required_permissions]

        permissions_list = [convert_string_to_permission(resource.resource_type, p) for p in required_permissions]
        required_permission = eval(resource.resource_type.capitalize() + "Permissions")(sum(permissions_list))

        if not compare_permissions(user_permissions, required_permission):
            actions = ", ".join([" ".join([j.capitalize() for j in i.split("_")]) for i in required_permissions])
            raise exceptions.ResourceExceptions.UserNotAuthorized(account,
                                                                  f"{resource.resource_type} {resource.id}",
                                                                  actions)



from . import account as AccountSchemas  # noqa: F401
from . import authentication as AuthSchemas  # noqa: F401
from . import group as GroupSchemas  # noqa: F401
from . import member as MemberSchemas  # noqa: F401
from . import policy as PolicySchemas  # noqa: F401
from . import poll as PollSchemas  # noqa: F401
from . import question as QuestionSchemas  # noqa: F401
from . import workspace as WorkspaceSchemas  # noqa: F401



from fastapi_users import schemas
from pydantic import ConfigDict, BaseModel, Field, EmailStr
from unipoll_api.documents import ResourceID


class Account(schemas.BaseUser[ResourceID]):
    id: ResourceID = Field(...)
    email: EmailStr = Field(...)
    first_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    last_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "email": "user@example.com",
            "password": "pass1234",
            "first_name": "John",
            "last_name": "Smith",
        }
    })


class AccountShort(BaseModel):
    id: ResourceID
    email: EmailStr
    first_name: str
    last_name: str


class AccountList(BaseModel):
    accounts: list[AccountShort]


class CreateAccount(schemas.BaseUserCreate):
    email: EmailStr = Field(...)
    first_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    last_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "email": "user@example.com",
            "password": "pass1234",
            "first_name": "John",
            "last_name": "Smith",
        }
    })


class UpdateAccount(schemas.BaseUserUpdate):
    email: EmailStr = Field(...)
    first_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    last_name: str = Field(
        default_factory=str,
        max_length=20,
        min_length=2,
        pattern="^[A-Z][a-z]*$")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "email": "user@example.com",
            "password": "pass1234",
            "first_name": "John",
            "last_name": "Smith",
        }
    })



from unipoll_api.schemas import PolicySchemas
from unipoll_api.utils import Permissions


# Get All Workspace Permissions
async def get_workspace_permissions() -> PolicySchemas.PermissionList:
    permissions = Permissions.WORKSPACE_ALL_PERMISSIONS.name.split('|')  # type: ignore
    return PolicySchemas.PermissionList(permissions=permissions)


# Get all possible group permissions
async def get_group_permissions() -> PolicySchemas.PermissionList:
    permissions = Permissions.GROUP_ALL_PERMISSIONS.name.split('|')  # type: ignore
    return PolicySchemas.PermissionList(permissions=permissions)



from unipoll_api.exceptions import resource
from unipoll_api.documents import ResourceID


# Exception for when a Policy is not found
class PolicyNotFound(resource.ResourceNotFound):
    def __init__(self, policy_id: ResourceID):
        super().__init__("Policy", resource_id=policy_id)


# Exception for when a PolicyHolder is not found
class PolicyHolderNotFound(resource.ResourceNotFound):
    def __init__(self, policy_holder_id: ResourceID):
        super().__init__("PolicyHolder", resource_id=policy_holder_id)



from unipoll_api.documents import ResourceID, Workspace, Account
from unipoll_api.exceptions import resource


# Exception for when a Workspace with the same name already exists
class NonUniqueName(resource.NonUniqueName):
    def __init__(self, workspace_name: str):
        super().__init__("Workspace", resource_name=workspace_name)


# Exception for when an error occurs during Workspace creation
class ErrorWhileCreating(resource.ErrorWhileCreating):
    def __init__(self, workspace_name: str):
        super().__init__("Workspace", resource_name=workspace_name)


# Exception for when a Workspace is not found
class WorkspaceNotFound(resource.ResourceNotFound):
    def __init__(self, workspace_id: ResourceID):
        super().__init__("Workspace", resource_id=workspace_id)


# Exception for trying to add a member that already exists
class AddingExistingMember(resource.AddingExistingMember):
    def __init__(self, workspace: Workspace, user: Account):
        super().__init__(workspace, user)


# Exception for when a Workspace was not deleted successfully
class ErrorWhileDeleting(resource.ErrorWhileDeleting):
    def __init__(self, workspace_id: ResourceID):
        super().__init__("Workspace", resource_id=workspace_id)


# Exception for when a user is not a member of the workspace
class UserNotMember(resource.UserNotMember):
    def __init__(self, workspace: Workspace, user: Account):
        super().__init__(workspace, user)


# Not authorized
class UserNotAuthorized(resource.UserNotAuthorized):
    def __init__(self, account: Account, workspace: Workspace, action: str = "perform this action in"):
        super().__init__(account, f"workspace {workspace.name}", action)


# Action not found
class ActionNotFound(resource.ActionNotFound):
    def __init__(self, action: str):
        super().__init__('workspace', action)


# Error while removing member
class ErrorWhileRemovingMember(resource.ErrorWhileRemovingMember):
    def __init__(self, workspace: Workspace, user: Account):
        super().__init__(workspace, user)



from beanie import WriteRules
from unipoll_api.documents import Poll, Workspace
from unipoll_api.schemas import PollSchemas, QuestionSchemas, WorkspaceSchemas
from unipoll_api.utils import Permissions
from unipoll_api.exceptions import ResourceExceptions, PollExceptions
from unipoll_api import actions


async def get_polls(workspace: Workspace | None = None,
                    check_permissions: bool = True) -> PollSchemas.PollList:
    all_workspaces = [workspace] if workspace else await Workspace.find(fetch_links=True).to_list()

    poll_list = []
    for workspace in all_workspaces:
        try:
            await Permissions.check_permissions(workspace, "get_polls", check_permissions)
            poll_list += workspace.polls  # type: ignore
        except ResourceExceptions.UserNotAuthorized:
            poll: Poll
            for poll in workspace.polls:  # type: ignore
                if poll.public:
                    poll_list.append(poll)
                else:
                    poll_list.append(await get_poll(poll, check_permissions))  # type: ignore
    # Build poll list and return the result
    for poll in poll_list:  # type: ignore
        poll_list.append(PollSchemas.PollShort(**poll.model_dump()))  # type: ignore
    return PollSchemas.PollList(polls=poll_list)


# Create a new poll in a workspace
async def create_poll(workspace: Workspace,
                      input_data: PollSchemas.CreatePollRequest,
                      check_permissions: bool = True) -> PollSchemas.PollResponse:
    # Check if the user has permission to create polls
    await Permissions.check_permissions(actions, "create_polls", check_permissions)

    # Check if poll name is unique
    poll: Poll  # For type hinting, until Link type is supported
    for poll in workspace.polls:  # type: ignore
        if poll.name == input_data.name:
            raise PollExceptions.NonUniqueName(poll)

    # Create a new poll
    new_poll = Poll(name=input_data.name,
                    description=input_data.description,
                    workspace=workspace,  # type: ignore
                    public=input_data.public,
                    published=input_data.published,
                    questions=input_data.questions,
                    policies=[])

    # Check if poll was created
    if not new_poll:
        raise PollExceptions.ErrorWhileCreating(new_poll)

    # Add the poll to the workspace
    workspace.polls.append(new_poll)  # type: ignore
    await Workspace.save(workspace, link_rule=WriteRules.WRITE)

    # Return the new poll
    return PollSchemas.PollResponse(**new_poll.model_dump())


async def get_poll(poll: Poll,
                   include_questions: bool = False,
                   include_policies: bool = False,
                   check_permissions: bool = True) -> PollSchemas.PollResponse:
    if not poll.public:
        await Permissions.check_permissions(poll, "get_poll", check_permissions)

    # Fetch the resources if the user has the required permissions
    questions = (await get_poll_questions(poll)).questions if include_questions else None
    policies = (await actions.PolicyActions.get_policies(resource=poll)).policies if include_policies else None

    workspace = WorkspaceSchemas.WorkspaceShort(**poll.workspace.model_dump())  # type: ignore

    # Return the workspace with the fetched resources
    return PollSchemas.PollResponse(id=poll.id,
                                    name=poll.name,
                                    description=poll.description,
                                    public=poll.public,
                                    published=poll.published,
                                    workspace=workspace,
                                    questions=questions,
                                    policies=policies)


async def get_poll_questions(poll: Poll,
                             check_permissions: bool = True) -> QuestionSchemas.QuestionList:
    # Check if the user has permission to get questions
    if not poll.public:
        await Permissions.check_permissions(poll, "get_questions", check_permissions)

    question_list = []
    for question in poll.questions:
        # question_data = question.model_dump()
        question_scheme = QuestionSchemas.Question(**question)
        question_list.append(question_scheme)
    # Return the list of questions
    return QuestionSchemas.QuestionList(questions=question_list)


async def update_poll(poll: Poll, data: PollSchemas.UpdatePollRequest) -> PollSchemas.PollResponse:
    # Update the poll
    if data.name:
        poll.name = data.name
    if data.description:
        poll.description = data.description
    if data.public is not None:
        poll.public = data.public
    if data.published is not None:
        poll.published = data.published
    if data.questions:
        poll.questions = data.questions

    # Save the updated poll
    await Poll.save(poll)
    return await get_poll(poll, include_questions=True)


async def delete_poll(poll: Poll):
    # Delete the poll
    await Poll.delete(poll)

    # Check if the poll was deleted
    if await Poll.get(poll.id):
        raise ResourceExceptions.InternalServerError("Poll not deleted")



from pydantic import BaseModel


class Question(BaseModel):
    id: int
    question: str
    question_type: str
    options: list[str]
    correct_answer: list[int]


class SingleChoiceQuestion(Question):
    question_type: str = "single-choice"


class MultipleChoiceQuestion(Question):
    question_type: str = "multiple-choice"


class OpenQuestion(Question):
    question_type: str = "open"


class QuestionList(BaseModel):
    questions: list[Question]



from fastapi import status
from unipoll_api.documents import Account, Resource, ResourceID
from unipoll_api.utils import Debug


class APIException(Exception):
    def __init__(self, code: int, detail: str):
        self.code = code
        self.detail = detail

    def __str__(self) -> str:
        Debug.print_error(self.detail)  # type: ignore
        return self.detail


class InternalServerError(APIException):
    def __init__(self, detail: str):
        super().__init__(code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                         detail="Internal Server Error")
        Debug.print_error(detail)  # type: ignore


class NonUniqueName(APIException):
    def __init__(self, resource: str, resource_name: str):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail=f"{resource} with name {resource_name} already exists")


class ErrorWhileCreating(APIException):
    def __init__(self, resource: str, resource_name: str):
        super().__init__(code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                         detail=f"Error while creating {resource} {resource_name}")


class ResourceNotFound(APIException):
    def __init__(self, resource: str, resource_id: ResourceID):
        super().__init__(code=status.HTTP_404_NOT_FOUND,
                         detail=f"{resource} #{resource_id} does not exist")


class ErrorWhileDeleting(APIException):
    def __init__(self, resource: str, resource_id: ResourceID):
        super().__init__(code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                         detail=f"Error while deleting {resource} #{resource_id}")


# Not authorized
class UserNotAuthorized(APIException):
    def __init__(self, account: Account, resource: str, action: str = "perform this action"):
        super().__init__(code=status.HTTP_403_FORBIDDEN,
                         detail=f"User {account.email} is not authorized to {action} in {resource}")


# Action not found
class ActionNotFound(APIException):
    def __init__(self, resource: str, action: str):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail=f"Action {action} not found in {resource}")


# Invalid permission
class InvalidPermission(APIException):
    def __init__(self, permission: str):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail=f"Invalid permission {permission}")


# User not a member of resource
class UserNotMember(APIException):
    def __init__(self, resource: Resource, user: Account):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail=f"User {user.email} is not a member of {resource.name} #{resource.id}")


# User already a member of resource
class AddingExistingMember(APIException):
    def __init__(self, resource: Resource, user: Account):
        super().__init__(code=status.HTTP_400_BAD_REQUEST,
                         detail=f"User {user.email} is already a member of {resource.name} #{resource.id}")


# Error while removing member
class ErrorWhileRemovingMember(APIException):
    def __init__(self, resource: Resource, user: Account):
        super().__init__(code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                         detail=f"Error while removing user {user.email} from {resource.name} #{resource.id}")



from unipoll_api import AccountManager
from unipoll_api import actions
from unipoll_api.documents import Workspace, Account, Policy
from unipoll_api.utils import Permissions
from unipoll_api.schemas import WorkspaceSchemas
from unipoll_api.exceptions import WorkspaceExceptions


# Get a list of workspaces where the account is a owner/member
async def get_workspaces(account: Account | None = None) -> WorkspaceSchemas.WorkspaceList:
    account = AccountManager.active_user.get()
    workspace_list = []

    search_result = await Workspace.find(Workspace.members.id == account.id).to_list()  # type: ignore

    # Create a workspace list for output schema using the search results
    for workspace in search_result:
        workspace_list.append(WorkspaceSchemas.WorkspaceShort(
            **workspace.model_dump(exclude={'members', 'groups', 'permissions'})))

    return WorkspaceSchemas.WorkspaceList(workspaces=workspace_list)


# Create a new workspace with account as the owner
async def create_workspace(input_data: WorkspaceSchemas.WorkspaceCreateInput) -> WorkspaceSchemas.WorkspaceCreateOutput:
    account: Account = AccountManager.active_user.get()
    # Check if workspace name is unique
    if await Workspace.find_one({"name": input_data.name}):
        raise WorkspaceExceptions.NonUniqueName(input_data.name)

    # Create a new workspace
    new_workspace = await Workspace(name=input_data.name, description=input_data.description).create()

    # Check if workspace was created
    if not new_workspace:
        raise WorkspaceExceptions.ErrorWhileCreating(input_data.name)

    await new_workspace.add_member(account=account, permissions=Permissions.WORKSPACE_ALL_PERMISSIONS)

    # Specify fields for output schema
    return WorkspaceSchemas.WorkspaceCreateOutput(**new_workspace.model_dump(include={'id', 'name', 'description'}))


# Get a workspace
async def get_workspace(workspace: Workspace,
                        include_groups: bool = False,
                        include_policies: bool = False,
                        include_members: bool = False,
                        include_polls: bool = False,
                        check_permissions: bool = True) -> WorkspaceSchemas.Workspace:
    await Permissions.check_permissions(workspace, "get_workspace", check_permissions)
    groups = (await actions.GroupActions.get_groups(workspace)).groups if include_groups else None
    members = (await actions.MembersActions.get_members(workspace)).members if include_members else None
    policies = (await actions.PolicyActions.get_policies(resource=workspace)).policies if include_policies else None
    polls = (await actions.PollActions.get_polls(workspace)).polls if include_polls else None
    # Return the workspace with the fetched resources
    return WorkspaceSchemas.Workspace(id=workspace.id,
                                      name=workspace.name,
                                      description=workspace.description,
                                      groups=groups,
                                      members=members,
                                      policies=policies,
                                      polls=polls)


# Update a workspace
async def update_workspace(workspace: Workspace,
                           input_data: WorkspaceSchemas.WorkspaceUpdateRequest,
                           check_permissions: bool = True) -> WorkspaceSchemas.Workspace:
    await Permissions.check_permissions(workspace, "update_workspace", check_permissions)
    save_changes = False

    # Check if user suplied a name
    if input_data.name and input_data.name != workspace.name:
        # Check if workspace name is unique
        if await Workspace.find_one({"name": input_data.name}) and workspace.name != input_data.name:
            raise WorkspaceExceptions.NonUniqueName(input_data.name)
        workspace.name = input_data.name  # Update the name
        save_changes = True
    # Check if user suplied a description
    if input_data.description and input_data.description != workspace.description:
        workspace.description = input_data.description  # Update the description
        save_changes = True
    # Save the updated workspace
    if save_changes:
        await Workspace.save(workspace)
    # Return the updated workspace
    return WorkspaceSchemas.Workspace(**workspace.model_dump())


# Delete a workspace
async def delete_workspace(workspace: Workspace, check_permissions: bool = True):
    await Permissions.check_permissions(workspace, "delete_workspace", check_permissions)

    # Delete all groups in the workspace
    for group in workspace.groups:
        await actions.GroupActions.delete_group(group)  # type: ignore

    await Workspace.delete(workspace)
    if await workspace.get(workspace.id):
        raise WorkspaceExceptions.ErrorWhileDeleting(workspace.id)
    await Policy.find({"parent_resource._id": workspace.id}, fetch_links=True).delete()



from unipoll_api.documents import Workspace
from unipoll_api.schemas import WorkspaceSchemas


# Get all workspaces
async def get_all_workspaces() -> WorkspaceSchemas.WorkspaceList:
    workspace_list = []
    search_result = await Workspace.find_all().to_list()

    # Create a workspace list for output schema using the search results
    for workspace in search_result:
        workspace_list.append(WorkspaceSchemas.Workspace(**workspace.model_dump()))

    return WorkspaceSchemas.WorkspaceList(workspaces=workspace_list)



from . import account_manager as AccountManager  # noqa: F401
from . import app as App  # noqa: F401
from . import config as Config  # noqa: F401
from . import dependencies as Dependencies  # noqa: F401
from . import documents as Documents  # noqa: F401
from . import mongo_db as MongoDB  # noqa: F401
from . import websocket_manager as WebsocketManager  # noqa: F401



from fastapi import Request
from unipoll_api.documents import ResourceID
from unipoll_api.utils.colored_dbg import print_error


# Extract resource from request
def extract_action_from_path(request: Request, extract_id: bool = True) -> str:
    # Extract path and operation_id from request
    try:
        return str(request["route"].name)
    except Exception as e:
        print_error("Error extracting resource from path: ", str(e))
        raise e


def extract_resourceID_from_path(request: Request) -> ResourceID:
    try:
        return ResourceID(request["path"].split("/")[2])
    except Exception as e:
        print_error("Error extracting resource ID from path: ", str(e))
        raise e



import motor.motor_asyncio  # type: ignore
from unipoll_api import documents as Documents
from unipoll_api.config import get_settings

settings = get_settings()

client = motor.motor_asyncio.AsyncIOMotorClient(
    host=settings.mongodb_url,
    uuidRepresentation="standard",
)

mainDB = client["unipoll-api"]

documentModels = [
    Documents.AccessToken,
    Documents.Resource,
    Documents.Account,
    Documents.Group,
    Documents.Workspace,
    Documents.Policy,
    Documents.Poll
]



from typing import Literal, Any, Optional
from pydantic import ConfigDict, BaseModel, Field
from unipoll_api.documents import ResourceID, Account, Group


class Policy(BaseModel):
    id: ResourceID
    policy_holder_type: Literal["account", "group"]
    policy_holder: Account | Group
    permissions: int


class PolicyShort(BaseModel):
    id: ResourceID
    policy_holder_type: Literal["account", "group"]
    policy_holder: Any = None
    permissions: Optional[Any] = None


class PolicyInput(BaseModel):
    policy_id: Optional[ResourceID] = Field(None, title="Policy ID")
    account_id: Optional[ResourceID] = Field(None, title="Account ID")
    group_id: Optional[ResourceID] = Field(None, title="Group ID")
    permissions: list[str] = Field(title="Permissions")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "permissions": ["get_workspace_info", "list_members"],
        }
    })


class PolicyOutput(BaseModel):
    permissions: list[str] = Field(title="List of allowed actions")
    policy_holder: Any = None
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "permissions": [
                "get_workspaces",
                "create_workspace",
                "get_workspace",
                "update_workspace",
                "delete_workspace",
                "get_workspace_members",
                "add_workspace_members",
                "remove_workspace_member",
                "get_groups",
                "create_group",
                "get_all_workspace_policies",
                "get_workspace_policy",
                "set_workspace_policy"
            ],
            "policy_holder": {
                "id": "1a2b3c4d5e6f7g8h9i0j",
                "email": "email@example.com",
                "first_name": "Name",
                "last_name": "Surname",
            }
        }
    })


# Schema for listing all policies in a workspace
class PolicyList(BaseModel):
    policies: list[PolicyShort] = Field(title="Policies")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "policies": [
                {
                    "permissions": [
                        "get_workspace",
                        "get_groups",
                    ],
                    "policy_holder": {
                            "id": "1a2b3c4d5e6f7g8h9i0j",
                            "email": "email@example.com",
                            "first_name": "Name",
                            "last_name": "Surname",
                    }
                },
                {
                    "permissions": [
                        "get_workspace",
                        "get_groups",
                    ],
                    "policy_holder": {
                            "id": "1a2b3c4d5e6f7g8h9i0j",
                            "email": "email@example.com",
                            "first_name": "Name",
                            "last_name": "Surname",
                    }
                }
            ]
        }
    })


# Schema for adding permissions to a group
class AddPermission(BaseModel):
    permissions: list[str] = Field(title="Permissions")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "permissions": [
                {
                    "type": "account",
                    "id": "1a2b3c4d5e6f7g8h9i0j",
                    "permission": "eff",
                },
                {
                    "type": "account",
                    "id": "2a3b4c5d6e7f8g9h0i1j",
                    "permission": "a3",
                },
                {
                    "type": "group",
                    "id": "3a4b5c6d7e8f9g0h1i2j",
                    "permission": "1",
                },
            ]
        }
    })


# Schema for returning a list of permissions
class PermissionList(BaseModel):
    permissions: list[str] = Field(title="Permissions")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "permissions": [
                "get_workspaces",
                "create_workspace",
                "get_workspace",
                "update_workspace",
                "delete_workspace",
                "get_workspace_members",
                "add_workspace_members",
                "remove_workspace_member",
                "get_groups",
                "create_group",
                "get_workspace_policies",
                "get_workspace_policy",
                "set_workspace_policy",
                "get_workspace_permissions"
            ]
        }
    })



from . import account as AccountExceptions  # noqa: F401
from . import authentication as AuthExceptions  # noqa: F401
from . import group as GroupExceptions  # noqa: F401
from . import policy as PolicyExceptions  # noqa: F401
from . import poll as PollExceptions  # noqa: F401
from . import resource as ResourceExceptions  # noqa: F401
from . import workspace as WorkspaceExceptions  # noqa: F401



from typing import Optional
from pydantic import ConfigDict, BaseModel, EmailStr, Field, root_validator
from unipoll_api.documents import ResourceID


# Schema for the response with basic member info
class Member(BaseModel):
    id: ResourceID
    email: Optional[EmailStr] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "id": "1a2b3c4d5e6f7g8h9i0j",
            "email": "user@example.com",
            "first_name": "John",
            "last_name": "Doe",
        }
    })


class AddMembers(BaseModel):
    accounts: list[ResourceID] = Field(title="Accounts")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "accounts": [
                "1a2b3c4d5e6f7g8h9i0j",
                "2a3b4c5d6e7f8g9h0i1j",
                "3a4b5c6d7e8f9g0h1i2j"
            ]
        }
    })


# Schema for the request to add a member to a workspace
class AddMembersRequest(BaseModel):
    accounts: list[ResourceID] = Field(title="Accounts")
    workspace: Optional[ResourceID] = Field(title="Workspace")
    group: Optional[ResourceID] = Field(title="Group")

    # Validate that either workspace or group is specified
    @root_validator(pre=True)
    def validate_resource(cls, values):
        if sum([bool(v) for v in values.values()]) != 2:
            raise ValueError('Either Workspace or Groups must be specified.')
        return values

    model_config = ConfigDict(validate_assignment=True, json_schema_extra={
        "example": {
            "accounts": [
                "1a2b3c4d5e6f7g8h9i0j",
                "2a3b4c5d6e7f8g9h0i1j",
                "3a4b5c6d7e8f9g0h1i2j"
            ]
        }
    },)


# Schema for the response with a list of members and their info
class MemberList(BaseModel):
    members: list[Member]
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "members": [
                {
                    "email": "jdoe@example.com",
                    "first_name": "John",
                    "last_name": "Doe",
                    "role": "admin"
                },
                {
                    "email": "jsmith@example.com",
                    "first_name": "Jack",
                    "last_name": "Smith",
                    "role": "user"
                }
            ]
        }
    })



from typing import Annotated
from functools import wraps
from fastapi import Cookie, Depends, Query, HTTPException, WebSocket
from unipoll_api.account_manager import active_user, get_current_active_user
from unipoll_api.documents import ResourceID, Workspace, Group, Account, Poll, Policy
from unipoll_api import exceptions as Exceptions


# Wrapper to handle exceptions and raise HTTPException
def http_dependency(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exceptions.ResourceExceptions.APIException as e:
            raise HTTPException(status_code=e.code, detail=str(e))
    return wrapper


# Dependency to get account by id
@http_dependency
async def get_account(account_id: ResourceID) -> Account:
    """
    Returns an account with the given id.
    """
    account = await Account.get(account_id)
    if not account:
        raise Exceptions.AccountExceptions.AccountNotFound(account_id)
    return account


async def websocket_auth(websocket: WebSocket,
                         session: Annotated[str | None, Cookie()] = None,
                         token: Annotated[str | None, Query()] = None) -> dict:
    return {"cookie": session, "token": token}


# Dependency for getting a workspace with the given id
@http_dependency
async def get_workspace(workspace_id: ResourceID) -> Workspace:
    """
    Returns a workspace with the given id.
    """
    workspace = await Workspace.get(workspace_id, fetch_links=True)

    if workspace:
        return workspace
    raise Exceptions.WorkspaceExceptions.WorkspaceNotFound(workspace_id)


# Dependency to get a group by id and verify it exists
@http_dependency
async def get_group(group_id: ResourceID) -> Group:
    """
    Returns a group with the given id.
    """
    group = await Group.get(group_id, fetch_links=True)
    if group:
        return group
    raise Exceptions.GroupExceptions.GroupNotFound(group_id)


# Dependency to get a poll by id and verify it exists
@http_dependency
async def get_poll(poll_id: ResourceID) -> Poll:
    """
    Returns a poll with the given id.
    """
    poll = await Poll.get(poll_id, fetch_links=True)
    if poll:
        return poll
    raise Exceptions.GroupExceptions.GroupNotFound(poll_id)


# Dependency to get a policy by id and verify it exists
@http_dependency
async def get_policy(policy_id: ResourceID) -> Policy:
    policy = await Policy.get(policy_id, fetch_links=True)
    if policy:
        # await policy.parent_resource.fetch_all_links()  # type: ignore
        return policy
    raise Exceptions.PolicyExceptions.PolicyNotFound(policy_id)


# Dependency to get a user by id and verify it exists
async def set_active_user(user_account: Account = Depends(get_current_active_user)):
    active_user.set(user_account)
    return user_account
